# vi: set softtabstop=2 ts=2 sw=2 expandtab:
# pylint:
"""
mergeconf - build a single configuration by merging multiple configuration
sources with order of precedence, based on immediacy.  Configurations can be
read from files (via ConfigParser), environment variables, and command-line
arguments (via ArgumentParser).  By default, configuration in files is
overridden by that in environment variables which in turn is overridden by
appropriate command-line arguments, but the library offers flexibility in
ordering.

## Examples

### Typical use

If the following is in `app.conf`:

```
shape = circle
upsidedown = false

[section2]
ratio = 20.403
count = 4
```

The following code could be used to set that up:

```
import mergeconf

conf = mergeconf.MergeConf('myapp', files='app.conf')
conf.add('name')
conf.add('shape', mandatory=True)
conf.add('colour', value='black')
conf.add('upsidedown', type=bool)
conf.add('rightsideup', type=bool, value=True)
section2 = conf.add_section('section2')
section2.add('count', type=int, mandatory=True)
section2.add('ratio', type=float)

# read file, override from environment, ensure mandatories are present
conf.merge()
```

Now to make use of the configuration:

```
# use attribute style access
print(f"Shape: {conf.shape}")

# including for sectioned configuration
print(f"Count: {conf.section2.count}")

# can also use array indices
print(f"Ratio: {conf['section2']['count']}")
```

### Handling atypical configuration hierarchy

In some cases it may be desirable to handle the merging yourself, such as if
you want a different hierarchy, such as environment configuration being
overridden by file-based configurations.

```
# not specifying file here
conf = mergeconf.MergeConf('myapp')
conf.add('name')
# other configuration items added, etc.
# ...

# now handle merge steps myself
conf.merge_file('app.conf')
conf.merge_environment()

# don't forget you have to validate when you're done
conf.validate()

# now ready to use
```

### Adding command-line arguments

MergeConf can work with Python's built-in `argparse` package to handle
command-line arguments in addition to configuration files and environment
variables:

```
conf = mergeconf.MergeConf('myapp')

# use the `cli` parameter to indicate inclusion in command-line arguments, and
# use the `description` parameter so there's help text in the command usage
conf.add('name', description='Name of the thing', cli=True)
...

# set up argparse however you like
parser = argparse.ArgumentParser(prog='myapp')
parser.add_argument(...)
...

# now call MergeConf to configure the parser for indicated configurations
conf.config_argparser(parser)

# parse arguments
args = parser.parse_args()

# merge configuration
conf.merge(args)
```

### Generating sample configuration

MergeConf also provides a way to generate a sample configuration file using the
configuration item descriptions, types and default values.

```
print(conf.sample_config())
```

Produces something like:

```
# (str) Unique name for the thing
#name =

# (str) The shape of the thing
shape =

# (str) The colour of the thing
#colour = black

# (bool) Upside-downness of the thing
#upsidedown =

# (bool) Is this thing right-side-up
#rightsideup = True

[section1]
# (str) What level of fluffiness does this item exhibit
#fluff = light

# (int) It's hard to come up with examples
#density =

[section2]
# (int) How many of the thing
count =

# (float) The ratio of thing to thang
#ratio =
```

Note that:
* expected types are described in the comment and what follows is the description
* default values are provided in the commented-out assignment
* optional items are commented out
* mandatory items are not commented out (but blank)
"""

from .mergeconf import MergeConf
from .mergeconfsection import MergeConfSection
from .mergeconfitem import MergeConfItem
from . import exceptions



# vi: set softtabstop=2 ts=2 sw=2 expandtab:
# pylint: disable=redefined-builtin,missing-module-docstring
# TODO: once `map` is retired, still can't remove redefined-builtin exception
#       due to use of `type` as a parameter name

from mergeconf.mergeconfitem import MergeConfItem

class MergeConfSection():
  """
  Container for configuration items and/or other sections.
  """
  def __init__(self, name, map=None):
    self._name = name
    self._items = {}
    self._sections = {}

    if map:
      for key, value in map.items():
        if isinstance(value, dict):
          self._sections[key] = MergeConfSection(key, map=value)
        else:
          self._items[key] = MergeConfItem(key, value)

  def __getitem__(self, key):
    if key in self._items:
      return self._items[key].value
    if key in self._sections:
      return self._sections[key]
    raise KeyError

  def __getattr__(self, attr):
    if attr in self._items:
      return self._items[attr].value
    if attr in self._sections:
      return self._sections[attr]
    raise AttributeError

  def __iter__(self):
    """
    Support iterating through configuration items.
    """
    for key, item in self._items.items():
      yield (key, item.value)

  # Given a list of variables with section prefix already stripped, find
  # variables defined in environment that match configuration items, and
  # assign their values to the items.  Then for each subsection, strip the
  # section prefix from any variables and call recursively on te section.
  def _merge_env(self, envvars):

    # get any variable names for this section
    for name, item in self._items.items():
      if name in envvars:
        item.value = envvars[name]

    # see if any subsection has variables
    for sectionname, section in self._sections.items():
      prefix = sectionname + '_'
      sectionvars = {
        # TODO(3.9): replace `split(prefix, 1)[1]` with `removeprefix(prefix)`
        x[0].split(prefix, 1)[1]: x[1]
        for x in envvars.items() if x[0].startswith(prefix)
      }
      if sectionvars:
        section._merge_env(sectionvars)

  def to_dict(self):
    """
    Return dictionary representation of configuration or section.
    """
    # TODO(3.9): use union
    # return {
    #   key: item.value for key, item in self._items.items()
    # } | {
    #   name: section.to_dict() for name, section in self._sections.items()
    # }
    d = { key: item.value for key, item in self._items.items() }
    d.update(
      { name: section.to_dict() for name, section in self._sections.items() }
    )
    return d

  @property
  def sections(self):
    """
    Return list of sections.
    """
    return self._sections.keys()

  def add(self, key, value=None, type=None, choices=None, mandatory=False,
      cli=False, description=None):
    """
    Add a configuration item.

    Args:
      key (str): Name of configuration item
      value (whatever): Default value, None by default
      type (type): Type of value
      choices (list): List of possible values
      mandatory (boolean): Whether item is mandatory or not, defaults to
        False.
      cli (boolean): Whether item should be included in command-line arguments
      description (str): Short descriptive text that may appear in usage text
        or sample configurations

    Notes: Type detection is attempted if not specified.
    """
    item = MergeConfItem(key, value, type=type, choices=choices,
      mandatory=mandatory, cli=cli, description=description)

    default = self._items.get(item.key, None)
    if default and not item.value:
      item.value = default.value
    self._items[item.key] = item

  def add_section(self, name):
    """
    Add a subsection to this section and return its object.
    """
    if name in self._sections:
      return self._sections[name]
    section = MergeConfSection(name)
    self._sections[name] = section
    return section

  def missing_mandatory(self):
    """
    Check that each mandatory item in this section and subsections has a
    defined value.

    Returns:
      List of fully qualified mandatory items without a defined value, in
      section-dot-item syntax.
    """
    def mandatories(sections, name, item):
      if item.mandatory and item.value is None:
        return f"{'.'.join(sections) + '.' if sections else ''}{name}"
      return None

    return self.map(mandatories) or None

  def _map(self, fn, sections):
    """
    Apply the given function to every item in this section and descend into
    subsections.

    Args:
      fn: Function taking (sections, name, MergeConfItem) and returning some
        value, or None.
      sections: list of sections built as a trail of breadcrumbs during
        recursion.

    Returns:
      List of values returned by function.  Values of None are not included.
    """
    results = []

    # apply to items
    for key, item in self._items.items():
      el = fn(sections, key, item)
      if el:
        results.append(fn(sections, key, item))

    # descend into subsections
    for name, section in self._sections.items():
      results.extend(section._map(fn, sections + [name]))

    return results

  def _sample_config(self):

    sample = []

    for name, item in self._items.items():
      typestr = f"({item.type.__name__}) " if item.type is not None else ''
      sample.append(f"# {typestr}{item.description or ''}")

      if item.choices:
        sample.append(f"# Can be one of: {', '.join(item.choices)}")

      if item.value:
        sample.append(f"#{name} = {item.value}")
      elif item.mandatory:
        sample.append(f"{name} =")
      else:
        sample.append(f"#{name} =")

      sample.append("")

    for name, section in self._sections.items():
      sample.append(f"[{name}]")
      sample.extend(section._sample_config())

    return sample



# vi: set softtabstop=2 ts=2 sw=2 expandtab:
# pylint: disable=redefined-builtin,invalid-name,missing-module-docstring
# TODO: Remove above exceptions.  At some point rename `type` parameter which
#       will allow us to remove redefined-builtin and invalid-name for
#       builtin_type

from mergeconf import exceptions

# aliasing this allows the use of a parameter `type`, for which I can't find a
# reasonable replacement (like `klass` for `class`)
builtin_type = type

class MergeConfItem:
  """
  Basic configuration item and base class for more complex types.
  """
  def _set_value_appropriately(self, value):
    if value is None:
      self._value = value
    elif self._choices is not None:
      if value in self._choices:
        self._value = self._type(value)
      else:
        # TODO: define the section :(
        raise exceptions.InvalidChoice("TODO: implement", self._key, self._choices, value)
    elif self._type == bool:
      if isinstance(value, bool):
        self._value = value
      else:
        self._value = value.lower() in ['true', 'yes', '1']
    else:
      self._value = self._type(value)

  def __init__(self, key, value=None, type=None, choices=None,
      mandatory=False, cli=False, description=None):
    """
    Create a configuration item.

    Arguments:
      key: Configuration item's key.
      value: Current value
      type: Item data type.  Must be one of bool, int, float or str.  If not
        specified, will be autodetected.
      choices: List of possible values.
      mandatory: Item must have configured value for configuration to be valid.
      cli: Include item in command-line argument parsing.
      description (str): Short descriptive text that may appear in usage text
        or sample configurations
    """
    if type and type not in [bool, int, float, str]:
      raise exceptions.UnsupportedType(type)
    if not type:
      if value is None:
        type = str
      else:
        type = builtin_type(value)
        if type not in [bool, int, float, str]:
          type = str

    self._key = key
    self._type = type
    self._choices = choices
    self._mandatory = mandatory
    self._cli = cli
    self._description = description

    self._set_value_appropriately(value)

  # pylint: disable=missing-function-docstring
  @property
  def key(self):
    return self._key

  @property
  def value(self):
    return self._value

  @value.setter
  def value(self, value):
    self._set_value_appropriately(value)

  @property
  def type(self):
    return self._type

  @property
  def choices(self):
    return self._choices

  @property
  def mandatory(self):
    return self._mandatory

  @property
  def cli(self):
    return self._cli

  @property
  def description(self):
    return self._description



# vi: set softtabstop=2 ts=2 sw=2 expandtab:
# pylint: disable=redefined-builtin
# TODO: at some point rename `type` parameter and remove exception for
#       redefined-builtin
"""
Exceptions raised by MergeConf package.
"""

class MissingConfiguration(Exception):
  """
  Raised if mandatory configuration items are missing.

  Attributes:
    missing: string list of missing items in section-dot-key notation,
      separated by commas.
  """

  def __init__(self, missing):
    self._missing = missing

    description = f"Undefined mandatory variables: {missing}"
    super().__init__(description)

  @property
  def missing(self):
    """ Return the string list of missing items """
    return self._missing

class MissingConfigurationFile(Exception):
  """
  Raised if the specified configuration file is missing or otherwise
  unreadable.

  Attributes:
    file: the missing file
  """

  def __init__(self, file):
    self._file = file
    description = f"Configuration file missing or unreadable: {file}"
    super().__init__(description)

  @property
  def file(self):
    """ Return the missing or unreadable file """
    return self._file

class UnsupportedType(Exception):
  """
  Raised if a configuration item is added with an unsupported type.

  Attributes:
    type: the unsupported type
  """

  def __init__(self, type):
    self._type = type.__name__
    description = f"Unsupported type: {self._type}"
    super().__init__(description)

  @property
  def type(self):
    """ Return the unsupported type """
    return self._type

class UndefinedSection(Exception):
  """
  Raised if a section is found that was not defined for the parser.

  Attributes:
    section: the section name
  """
  def __init__(self, section):
    self._section = section
    description = f"Unexpected section found: '{section}'"
    super().__init__(description)

  @property
  def section(self):
    """ Return the undefined section """
    return self._section

class UndefinedConfiguration(Exception):
  """
  Raised if a configuration item is found that was not defined for the parser.

  Attributes:
    section: the section name
    item: the item name
  """
  def __init__(self, section, item):
    self._section = section
    self._item = item
    description = f"Unexpected configuration item found: '{item}' in section '{section}'"
    super().__init__(description)

  @property
  def section(self):
    """ Return the section where the item was found """
    return self._section

  @property
  def item(self):
    """ Return the undefined item """
    return self._item

class InvalidChoice(Exception):
  """
  Raised when a value is provided for a configuration item with set choices,
  and is not of those choices.

  Attributes:
    section: the section name
    item: the item name
    choices: valid choices defined for the item
    value: value provided for the item
  """
  def __init__(self, section, item, choices, value):
    self._section = section
    self._item = item
    self._choices = choices
    self._value = value
    super().__init__(
      f"Invalid configuration value '{value}' provided for {section}.{item} (not in {choices})"
    )

  @property
  def section(self):
    """ Return the configuration section """
    return self._section

  @property
  def item(self):
    """ Return the configuration item """
    return self._item

  @property
  def choices(self):
    """ Return the valid choices """
    return self._choices

  @property
  def value(self):
    """ Return the value given """
    return self._value

class AlreadyMerged(Exception):
  """
  Raised when client attempts to define configuration after merging.
  """

class Unmerged(Exception):
  """
  Raised when the configuration is accessed before merging.
  """

#class Deprecated(Exception):
#  """
#  Raised for hard deprecations where functionality has been removed and the
#  API is not available at all.
#
#  Attributes:
#    version: the last version in which this functionality is available.
#    message: further information to assist the user.
#  """
#  def __init__(self, version, message=None):
#    self._version = version
#    self._message = message
#    self._function = inspect.stack()[1].function
#    etc = f": {message}" if message else '.'
#    description = f"Deprecated API `{self._function}` last available in version {version}{etc}"
#    super().__init__(description)
#
#  @property
#  def function(self):
#    return self._function
#
#  @property
#  def version(self):
#    return self._version



# vi: set softtabstop=2 ts=2 sw=2 expandtab:
# pylint: disable=redefined-builtin,missing-module-docstring
# TODO: once `map` is retired, remove redefined-builtin exception
import os
import logging
from configparser import ConfigParser
from mergeconf import exceptions
from mergeconf.mergeconfsection import MergeConfSection


class MergeConf(MergeConfSection):
  """
  Configuration class.  Initialized optionally with configuration items, then
  additional items may be added explicitly (and must be if they are mandatory,
  a specific type, etc.).  Once all items have been added the configuration is
  finalized with merge(), validation checks are performed, and the realized
  values can be extracted.

  This class inherits from the MergeConfSection class, which contains methods
  to define configuration items and sections and examine the configuration.
  """

  def __init__(self, codename, files=None, map=None, strict=True):
    """
    Initializes MergeConf class.

    Args:
      codename (str): Simple string which is assumed to prefix any related
        environment variables associated with the configuration (along with an
        underscore as separator), in order to avoid collisions in the
        environment's namespace.  For example, for an `app_name` configuration
        key, with a codename `MYAPP`, the corresponding environment variable
        would be `MYAPP_APP_NAME`.
      files (str or list): filename or list of filenames for configuration
        files.  Files are applied in order listed, and so should be listed from
        least to most important.
      map (dict): Configuration options which are neither mandatory nor of a
        specified type, specified as key, value pairs.
      strict (boolean): If true, unexpected configuration sections or items
        will cause an exception (`UndefinedSection` or `UndefinedConfiguration`,
        respectively).  If false, they will be added to the merged
        configuration.

    Note: The `map` argument is probably to be deprecated and removed at a
      later date.  Its utility is limited and should be avoided.
    """
    super().__init__(None, map=map)

    self._args = None
    self._argparser = None
    self._codename = codename
    self._strict = strict
    self._merged = False

    # turn given files parameter into an iterable sequence if not already
    self._files = files
    if files and not isinstance(files, (list, tuple)):
      self._files = (files,)

    # main section name transparently added.  ConfigParser requires all items
    # to be contained in a section; this supports simpler configurations and
    # avoids having to create a "main" or "app" section explicitly if not
    # desired.
    self._main = '__app__'

    if map:
      logging.warning("Support for `map` argument is deprecated and will " \
        "be removed.  Please use `add()` to add configuration options and " \
        "their specifications, including default values.")

  # overload this here in order to try and catch non-mergeconf CLI arguments
  # if appropriate
  def __getattr__(self, attr):
    if not self._merged:
      raise exceptions.Unmerged
    try:
      return super().__getattr__(attr)
    except AttributeError as e:
      if self._args is not None and attr in vars(self._args):
        return vars(self._args)[attr]
      raise e

  def map(self, fn):
    """
    Apply the given function to every item in this section and recursively for
    subsections.

    Args:
      fn: Function taking (sections, name, MergeConfItem) and returning some
        value, or None.

    Returns:
      List of values returned by function.  Values of None are not included.
    """
    return self._map(fn, [])

  def config_argparser(self, argparser):
    """
    Configure ArgumentParser instance with designated configuration items.

    This will run through all configuration items and add any defined as
    appropriate for command-line arguments in the parser.  This method must
    therefore be called before the ArgumentParser instance can be used.  The
    client may configure any of its own arguments and other sections before
    and/or after calling this method.

    Arguments are configured with help text based on the configuration items'
    descriptions, if available.  Boolean configuration items do not take
    arguments but instead will set a value opposite of their default, or True
    if not was defined.

    args:
      argparser: ArgumentParser object to populate with appropriate items.
    """
    def addargs(sections, name, item):
      if item.cli:
        section_prefix = '-'.join(sections) + '-' if sections else ''
        argname = f"--{section_prefix}{name.replace('_', '-')}"
        kwargs = {
          'help': item.description,
        }
        if item.type == bool:
          # the default action for a boolean should be store_true
          kwargs['action'] = 'store_false' if item.value is True else 'store_true'
        else:
          kwargs['type'] = item.type
          if item.choices is not None:
            kwargs['choices'] = item.choices
          else:
            kwargs['metavar'] = name.upper()
        argparser.add_argument(argname, **kwargs)

    # can't add argparser after configuration is merged
    if self._merged:
      raise exceptions.AlreadyMerged

    self._argparser = argparser
    self.map(addargs)

  def merge_args(self, args):
    """
    Merge command-line arguments parsed by ArgumentParser.

    Only configuration items identified with `cli=True` on creation
    (in `add()`) will be considered.  See `config_argparser()` for adding the
    arguments to an ArgumentParser object automatically.

    Args:
      args: Arguments returned by parse_args().
    """
    argsd = vars(args)

    def grokarg(sections, name, item):
      if item.cli:
        argname = f"{'_'.join(sections) + '_' if sections else ''}{name}"
        if argsd.get(argname) is not None:
          item.value = argsd[argname]

    self.map(grokarg)

    # retain args for retrieving individual non-mergeconf CLI args
    self._args = args

  def merge_environment(self):
    """
    Using configuration definition, reads in variables from the environment
    matching the pattern `<codename>[_<section_name>]_<variable_name>`.  Any
    variable found not matching a defined configuration item is returned in
    a list: in this way variables outside the merged configuration context can
    be handled, such as a variable specifying an alternative config file.

    Returns:
      Map of environment variables matching the application codename.  The
      keys will be stripped of the codename prefix and will be converted to
      lowercase.
    """
    # add this to any environment variable names
    prefix = self._codename.upper() + '_'

    # get all environment variables starting with that prefix into dict with
    # key stripped of prefix and made lowercase
    envvars = {
      # TODO(3.9): replace `split(prefix, 1)[1]` with `removeprefix(prefix)`
      x[0].split(prefix, 1)[1].lower(): x[1]
      for x in os.environ.items() if x[0].startswith(prefix)
    }

    self._merge_env(envvars)

    return envvars

  def merge_file(self, config_file):
    """
    Merge configuration defined in file.  File is expected to adhere to the
    format defined by ConfigParser, with `=` used as the delimiter and
    interpolation turned off.  In addition, unlike ConfigParser, config files
    may include variables defined prior to any section header.

    Args:
      config_file (str): Path to config file.
    """
    config = ConfigParser(delimiters='=', interpolation=None)

    # read configuration into string so we can prepend a pretend main section.
    # See definition of self._main for explanation.
    try:
      with open(config_file, encoding="utf-8") as f:
        config_content = f"[{self._main}]\n{f.read()}"
    except FileNotFoundError:
      # pylint: disable=raise-missing-from
      raise exceptions.MissingConfigurationFile(config_file)

    # read configuration
    config.read_string(config_content, source=config_file)

    # read into stuffs
    for section in config.sections():
      if section == self._main:
        ref = self
      elif section not in self._sections:
        # unrecognized configuration section
        if self._strict:
          raise exceptions.UndefinedSection(section)
        logging.warning("Unexpected section in configuration: %s", section)
        ref = self.add_section(section)
      else:
        ref = self._sections[section]
      for option in config.options(section):
        if option not in ref._items:
          if self._strict:
            raise exceptions.UndefinedConfiguration(section, option)
          logging.warning("Unexpected configuration item in section %s: %s",
            section, option)
          ref.add(option, config[section][option])
        else:
          ref._items[option].value = config[section][option]

  def validate(self):
    """
    Checks that mandatory items have been defined in configuration.  If not,
    throws exception.  Client may also use `missing_mandatory()`.

    Subclasses may add additional validation but should first call the parent
    implementation as the test for mandatory items is primary.
    """
    # TODO(3.8): use walrus operator
    # if unfulfilled := self.missing_mandatory():
    unfulfilled = self.missing_mandatory()
    if unfulfilled:
      raise exceptions.MissingConfiguration(', '.join(unfulfilled))

    # we can now consider the configuration merged
    self._merged = True

  def merge(self, args=None, unparsed=None):
    """
    Takes configuration definition and any configuration files specified and
    reads in configuration, overriding default values.  These are in turn
    overridden by corresponding variables found in the environment, if any.
    Basic validations are performed.

    This is a convenience method to handle the typical configuration
    hierarchy and process.  Clients may also call other `merge_*` methods in
    any order, but should call `validate()` if so to ensure all mandatory
    configurations are specified.

    Args:
      args: Arguments processed by ArgumentParser.  Any matching appropriate
        are merged in after environment variables.
      unparsed (list of strings): Unparsed arguments.
    """

    # get configuration file(s) from environment, fall back to default
    from_env = os.environ.get(f"{self._codename.upper()}_CONFIG")
    config_files = from_env.split(',') if from_env else self._files

    # if we have config files, merge into config
    if config_files:
      for config_file in config_files:
        logging.debug("Merging in config file %s", config_file)
        self.merge_file(config_file)

    # override with variables set in environment
    self.merge_environment()

    # override further with command-line arguments, if available
    if args:
      self.merge_args(args)
    elif self._argparser is not None:
      self._args = self._argparser.parse_args(args=unparsed)
      self.merge_args(self._args)

    # test that mandatory values have been set
    self.validate()

  def sample_config(self):
    """
    Create a sample configuration.

    This will be more informative if configuration items have been specified
    with descriptions.

    Returns:
      A string describing a sample configuration file.

    Note:
      The sample configuration will have this format:

      ```
      # (str) this is the first item
      name =

      # (int) this is the second item which has a default value
      #count = 1

      [section1]
      # (bool) this item has no default
      #has_car =

      # (str) This is mandatory
      description =
      ```
    """
    # The internal method for building sample config will always append an
    # extra empty string.  Get rid of this before joining all the lines
    # together with a newline to build the "file".
    return '\n'.join(self._sample_config()[:-1])

  # TODO: Not sure if it's a good model to provide this.  Flexibility is good,
  #       but while we need this internally, doing everything needed with the
  #       parser before merging the config and not touching it afterwards
  #       in favour of consuming the config and the args seems correct.
  #@property
  #def argparser(self):
  #  return self._argparser

  @property
  def args(self):
    """ Returns parsed arguments.
    """
    if not self._merged:
      raise exceptions.Unmerged()
    return self._args



# vi: set softtabstop=2 ts=2 sw=2 expandtab:
# pylint: disable=unused-import,singleton-comparison,redefined-outer-name
"""
Basic operation tests.
"""
import os
import pytest
from tests.fixtures import (
  config, config_no_file, config_with_defaults, config_not_strict,
  config_strict, argparser, CODENAME
)
import mergeconf

# ---------------------------------------------------------------------------
#                                                             helpers
# ---------------------------------------------------------------------------

CODENAMEUC = CODENAME.upper()

def envvarname(without_prefix):
  """ Create environment variable name """
  return f"{CODENAMEUC}_{without_prefix.upper()}"

def clean_up_env():
  """ Clear environment variables previously set in testing """
  for key in os.environ:
    if key.startswith(CODENAMEUC):
      del os.environ[key]

# ---------------------------------------------------------------------------
#                                                             tests
# ---------------------------------------------------------------------------

def test_no_config(config_no_file):
  """
  Tests we are able to parse an empty configuration, except it will fail on
  missing mandatory field ("shape").
  """
  with pytest.raises(mergeconf.exceptions.MissingConfiguration) as e:
    config_no_file.merge()
  print(e.value.missing)
  assert e.value.missing == "shape, section2.count"

def test_no_config_map(config_with_defaults):
  """
  Tests we are able to parse an empty configuration and get a map back.
  """
  config_with_defaults.merge()
  assert config_with_defaults['shape'] == 'triangle'
  assert config_with_defaults.shape == 'triangle'

  d = config_with_defaults.to_dict()
  print("Config as dict:")
  print(d)
  assert d == {
    'name': None,
    'colour': 'blue',
    'upsidedown': None,
    'rightsideup': True,
    'shape': 'triangle',
    'section2': {
      'count': 13,
      'z_index': 12,
      'ratio': None,
      'ferbs': '[1, 2, 3, 4]'
    }
  }

def test_config_only_env(config):
  """
  Tests we are able to parse a config defined only in the environment.
  """
  os.environ[envvarname("SHAPE")] = 'circle'
  os.environ[envvarname("UPSIDEDOWN")] = 'no'
  os.environ[envvarname("SECTION2_COUNT")] = '10'
  os.environ[envvarname("SECTION2_RATIO")] = '10.425'
  os.environ[envvarname("SECTION2_FUBAR")] = 'this does not even matter'
  config.merge()
  assert config['shape'] == 'circle'
  assert config['upsidedown'] == False
  assert config['rightsideup'] == True
  assert config['section2']['count'] == 10
  assert config['section2']['ratio'] == 10.425
  assert config.shape == 'circle'
  assert config.upsidedown == False
  assert config.rightsideup == True
  assert config.texture == 'smooth'
  assert config.section2.count == 10
  assert config.section2.ratio == 10.425

  # clean up environment
  clean_up_env()

def test_config_only_file(config):
  """
  Tests we are able to correctly parse a config defined only in a file.
  """
  config.merge()
  assert config['name'] is None
  assert config['shape'] == 'circle'
  assert config['upsidedown'] == False
  assert config['rightsideup'] == True
  assert config['texture'] == 'smooth'
  assert config['section2']['count'] == 4
  assert config['section2']['ratio'] == 20.403

def test_config_file_and_env(config):
  """
  Tests we are able to correctly parse a config defined in both file and
  in the environment, with environment taking precedence.
  """
  os.environ[envvarname("SHAPE")] = 'triangle'
  os.environ[envvarname("UPSIDEDOWN")] = 'true'
  os.environ[envvarname("SECTION2_COUNT")] = '15'
  config.merge()
  assert config['shape'] == 'triangle'
  assert config['upsidedown'] == True
  assert config['rightsideup'] == True
  assert config['section2']['count'] == 15
  assert config['section2']['ratio'] == 20.403

  # clean up environment
  clean_up_env()

def test_config_file_and_args(config, argparser):
  """
  Tests we are able to correctly parse a config defined in both file and
  in arguments, with arguments taking precedence.
  """
  # configure argparser to support the designated config items in the fixture
  config.config_argparser(argparser)

  args = argparser.parse_args(['--upsidedown', '--section2-ratio=2.22'])
  config.merge(args)

  # test values are as expected
  assert config['shape'] == 'circle'
  assert config['rightsideup'] == True
  assert config['section2']['count'] == 4
  assert config['section2']['ratio'] == 2.22
  assert config['texture'] == 'smooth'

def test_config_missing_file(config):
  """
  Tests we handle a missing config file.
  """
  with pytest.raises(mergeconf.exceptions.MissingConfigurationFile) as e:
    config.merge_file('test2_missing.conf')
  assert e
  assert e.value.file == 'test2_missing.conf'

  # clean up environment
  clean_up_env()

def test_unsupported_type():
  """
  Tests the attempted use of an unsupported type throws an exception.
  """
  conf = mergeconf.MergeConf('test')
  with pytest.raises(mergeconf.exceptions.UnsupportedType) as e:
    conf.add('SECTION1_NAME', type=list)
  assert e.value.type == 'list'

def test_default_values(config_with_defaults):
  """
  Tests a configuration initialized with defaults via a map and merging in
  a file.
  """
  config_with_defaults.merge_file('tests/test2.conf')
  assert config_with_defaults['shape'] == 'rectangle'
  assert config_with_defaults['colour'] == 'blue'
  assert config_with_defaults['upsidedown'] == False
  assert config_with_defaults['rightsideup'] == True
  assert config_with_defaults['section2']['count'] == 10
  assert config_with_defaults['section2']['ratio'] == 20.403
  assert config_with_defaults['section2']['z_index'] == 12

def test_add_with_defaults(config_with_defaults):
  """
  Tests a configuration initialized with defaults via a map.
  """
  section2 = config_with_defaults.add_section('section2')
  section2.add('feels', value='heavy')
  section2.add('width', value=10)
  section2.add('transparent', value=False)
  section2.add('ferbity', value=4.2)
  section2.add('ferbs', value=[1, 2, 3, 4])
  config_with_defaults.merge()

  # basically ensure none of the above become strings unless they already
  # are, or are an unsupported type
  assert config_with_defaults['section2']['feels'] == 'heavy'
  assert config_with_defaults['section2']['width'] == 10
  assert config_with_defaults['section2']['transparent'] == False
  assert config_with_defaults['section2']['ferbity'] == 4.2
  assert config_with_defaults['section2']['ferbs'] == '[1, 2, 3, 4]'
  assert config_with_defaults.section2.ferbs == '[1, 2, 3, 4]'

def test_iterate_values(config_with_defaults):
  """
  Tests iteration through a configuration.
  """
  config_with_defaults.merge_file('tests/test2.conf')
  print(config_with_defaults.to_dict())

  # iterate through main items
  it = iter(config_with_defaults)
  assert next(it)[0] == 'colour'
  assert next(it)[0] == 'shape'
  assert next(it)[0] == 'name'
  assert next(it)[0] == 'upsidedown'
  assert next(it)[0] == 'rightsideup'

  # iterate through sections
  for name in config_with_defaults.sections:
    assert name == 'section2'
    section = config_with_defaults[name]
    it = iter(section)
    assert next(it)[0] == 'count'
    assert next(it)[0] == 'z_index'
    assert next(it)[0] == 'ferbs'
    assert next(it)[0] == 'ratio'

def test_unconfigured_section_allowed(config_not_strict):
  """
  Tests that an unconfigured section is allowed when `strict` isn't enabled.
  """
  config_not_strict.merge_file('tests/test3.conf')
  config_not_strict.validate()
  assert config_not_strict['section2']['snurf'] == 'garbage'
  assert config_not_strict['section3']['blarb'] == '32'
  assert config_not_strict['section3']['snarf'] == 'this is the way the world ends'

def test_unconfigured_item_not_allowed(config_strict):
  """
  Tests that an unconfigured configuration item is caught and an exception thrown.
  """
  with pytest.raises(mergeconf.exceptions.UndefinedConfiguration) as e:
    config_strict.merge_file('tests/test3.conf')
    config_strict.validate()
  assert e.value.section == 'section2'
  assert e.value.item == 'snurf'

def test_unconfigured_section_not_allowed(config_strict):
  """
  Tests that an unconfigured configuration section is caught and an exception thrown.
  """
  with pytest.raises(mergeconf.exceptions.UndefinedSection) as e:
    config_strict.merge_file('tests/test4.conf')
    config_strict.validate()
  assert e.value.section == 'section3'

#def test_deprecated_functionality(config):
#  """
#  Tests that attempted use of deprecated functionality results in an exception.
#  """
#  with pytest.raises(mergeconf.exceptions.Deprecated) as e:
#    config.add_boolean('thisshouldfail', True)
#  assert e.value.function == 'add_boolean'
#  assert e.value.version == '0.3'
#  with pytest.raises(mergeconf.exceptions.Deprecated) as e:
#    config.parse('fakefilename')
#  assert e.value.function == 'parse'
#  assert e.value.version == '0.3'

def test_garbage_item_by_index(config):
  """
  Tests that attempted access of undefined variable results in an exception.
  """
  config.merge()
  with pytest.raises(KeyError):
    print(config['whut'])

def test_garbage_item_by_attribute(config):
  """
  Tests that attempted access of undefined variable results in an exception.
  """
  config.merge()
  with pytest.raises(AttributeError):
    print(config.whut)

def test_map(config):
  """
  Tests the map() function by building a list of mandatory configuration
  items.
  """
  config.merge()

  def mandatories(sections, name, item):
    if item.mandatory:
      return f"{'.'.join(sections) + '.' if sections else ''}{name}"
    return None

  res = config.map(mandatories)
  print(res)
  assert res == ['shape', 'section2.count']

def test_args(config, argparser):
  """
  Tests that argument parsing is set up appropriately.
  """
  # configure argparser to support the designated config items in the fixture
  config.config_argparser(argparser)

  # now try interpreting some stuff
  argparser.parse_args(['--shape=square', '--upsidedown', '--section2-count=12'])
  assert argparser.format_usage() == \
    """usage: test [-h] [-c CONFIG] [-d] [-q] [--shape SHAPE] [--colour COLOUR]
            [--upsidedown] [--rightsideup]
            [--texture {smooth,furry,gelatinous}] [--section1-fluff FLUFF]
            [--section2-count COUNT] [--section2-ratio RATIO]
"""
  config.merge(unparsed=[
    '--shape=square', '--upsidedown', '--texture=furry', '--section2-count=12'
  ])
  assert config.shape == 'square'
  assert config.section1.fluff == 'light'
  assert config.section2.count == 12
  assert config.upsidedown is True
  assert config.rightsideup is True
  assert config.texture == 'furry'
  assert config.debug is False
  with pytest.raises(AttributeError):
    assert config.whatnow is None

def test_args_explicit(config, argparser):
  """
  Tests that argument parsing is set up appropriately, using the older method
  of passing in the args explicitly.
  """
  # configure argparser to support the designated config items in the fixture
  config.config_argparser(argparser)

  # now try interpreting some stuff
  args = argparser.parse_args(['--shape=square', '--upsidedown', '--section2-count=12'])
  assert argparser.format_usage() == \
    """usage: test [-h] [-c CONFIG] [-d] [-q] [--shape SHAPE] [--colour COLOUR]
            [--upsidedown] [--rightsideup]
            [--texture {smooth,furry,gelatinous}] [--section1-fluff FLUFF]
            [--section2-count COUNT] [--section2-ratio RATIO]
"""
  config.merge(args)
  assert config.shape == 'square'
  assert config.section1.fluff == 'light'
  assert config.section2.count == 12
  assert config.upsidedown is True
  assert config.rightsideup is True
  assert config.debug is False
  with pytest.raises(AttributeError):
    assert config.whatnow is None

def test_unmerged(config, argparser):
  """
  Tests an exception is thrown if configuration isn't merged but we try to get
  at the configuration.
  """
  config.config_argparser(argparser)
  argparser.parse_args(['--shape=square', '--upsidedown', '--section2-count=12'])
  with pytest.raises(mergeconf.exceptions.Unmerged):
    assert config.shape == 'square'
  with pytest.raises(mergeconf.exceptions.Unmerged):
    assert config.args.shape == 'square'

def test_already_merged(config, argparser):
  """
  Tests an exception is thrown if configuration is merged and we try to update
  the configuration.
  """
  config.merge()
  with pytest.raises(mergeconf.exceptions.AlreadyMerged):
    config.config_argparser(argparser)

def test_args_mandatory_in_config(config, argparser):
  """
  Tests that argument parsing succeeds even if a mandatory item is only
  defined in the configuration.
  See https://gitlab.com/dleske/mergeconf/-/issues/2
  """
  # configure argparser to support the designated config items in the fixture
  config.config_argparser(argparser)

  # now try interpreting some stuff
  assert argparser.format_usage() == \
    """usage: test [-h] [-c CONFIG] [-d] [-q] [--shape SHAPE] [--colour COLOUR]
            [--upsidedown] [--rightsideup]
            [--texture {smooth,furry,gelatinous}] [--section1-fluff FLUFF]
            [--section2-count COUNT] [--section2-ratio RATIO]
"""
  config.merge(unparsed=['--section2-count=12'])
  assert config.shape == 'circle'
  assert config.section1.fluff == 'light'
  assert config.section2.count == 12
  assert config.upsidedown is False
  assert config.rightsideup is True
  assert config.debug is False

  # I mean don't do it this way but you could
  assert config.args.section2_count == 12

  with pytest.raises(AttributeError):
    assert config.whatnow is None

def test_items_with_underscores(config, argparser):
  """
  Test that items with underscores in the name are accepted and handled
  correctly.
  """
  section2 = config.add_section('section2')
  section2.add('feels_weight', value='heavy', cli=True)
  section2.add('feels_texture', value='smooth', cli=True)

  # configure argparser to support the designated config items in the fixture
  config.config_argparser(argparser)

  # now try interpreting some stuff
  # TODO: can we make the metavars something else somehow
  assert argparser.format_usage() == \
    """usage: test [-h] [-c CONFIG] [-d] [-q] [--shape SHAPE] [--colour COLOUR]
            [--upsidedown] [--rightsideup]
            [--texture {smooth,furry,gelatinous}] [--section1-fluff FLUFF]
            [--section2-count COUNT] [--section2-ratio RATIO]
            [--section2-feels-weight FEELS_WEIGHT]
            [--section2-feels-texture FEELS_TEXTURE]
"""
  config.merge(unparsed=['--section2-feels-texture=rough'])

  # basically ensure none of the above become strings unless they already
  # are, or are an unsupported type
  assert config['section2']['feels_weight'] == 'heavy'
  assert config['section2']['feels_texture'] == 'rough'

def test_bad_choices(config, argparser):
  """
  Test that items with invalid choices are rejected correctly.
  """
  # configure argparser to support the designated config items in the fixture
  config.config_argparser(argparser)

  # try a configuration which tries to define an invalid choice
  with pytest.raises(SystemExit):
    argparser.parse_args(['--texture=sandy'])

def test_bad_choices_env(config):
  """
  Test that bad choices defined in environment are rejected correctly.
  """
  os.environ[envvarname("TEXTURE")] = 'sorta like, fudgey?'
  with pytest.raises(mergeconf.exceptions.InvalidChoice) as e:
    config.merge()
    # assert exception is like damn
  assert e.value.section == 'TODO: implement'
  assert e.value.item == 'texture'
  assert e.value.choices == ['smooth', 'furry', 'gelatinous']
  assert e.value.value == 'sorta like, fudgey?'

  # clean up environment
  clean_up_env()

def test_sample_config(config):
  """
  Test that the sample configuration is generated correctly.
  """
  sample_config = config.sample_config()
  print("Sample configuration:\n\n```")
  print(sample_config)
  print("```")
  assert sample_config == \
"""# (str) Unique name for the thing
#name =

# (str) The shape of the thing
shape =

# (str) The colour of the thing
#colour = black

# (bool) Upside-downness of the thing
#upsidedown =

# (bool) Is this thing right-side-up
#rightsideup = True

# (str) The texture the thing exhibits (passively)
# Can be one of: smooth, furry, gelatinous
#texture =

[section1]
# (str) What level of fluffiness does this item exhibit
#fluff = light

# (int) It's hard to come up with examples
#density =

[section2]
# (int) How many of the thing
count =

# (float) The ratio of thing to thang
#ratio ="""
